@@ -54,7 +54,6 @@ class CloudCostAllocator(object):
         for cost_item in cost_items:
             if date_str != cost_item.date_str:
                 error("Found cost items with different dates: " + date_str + " and " + cost_item.date_str)
-                return False
 
         # Check all currencies are the same and populate missing ones (currencies are missing for consumer
         # cost items)
@@ -173,7 +172,7 @@ class CloudCostAllocator(object):
                                              cost_items: list[CostItem]) -> None:
 
         # Check if consumer tag is used and process cost items
-        if 'ConsumerService' in self.config['TagKey']:
+        if 'ConsumerService' in self.config['TagKey'] or 'Product' in self.config['TagKey']:
             for cost_item in cost_items:
 
                 # Check consumer service
@@ -200,6 +199,7 @@ class CloudCostAllocator(object):
                             break
 
                 # Check product
+                # TODO: product dimension
                 product = ""
                 actual_product_tag_key = ""
                 if 'Product' in self.config['TagKey']:
